# World Library Comparison.
from _thread import start_new_thread as nth
from time import time as now
from datetime import datetime
from os.path import splitext, basename
from optparse import OptionParser
from pickle import load as unpickle, dump as pickle
from sys import exc_info

from stuphos import enqueueHeartbeatTask
from stuphos.etc.tools.timing import Elapsed
from stuphos.etc.tools import parseOptionsOverSystem
from stuphmud.server.player import HandleCommandError
from stuphmud.server.player.shell import PromptFor

from stuphlib.wldlib import World
from stuphlib.dblib import LocalIndex, ArchiveIndex, ZipArchiveIndex

from stuphmud.server.player.db.managedfs import getCheckedPlayerFilename, PERSONAL_PLRFILE_DIR
from .wizard import parameters

# Preparation.
def nowstr():
    return datetime.fromtimestamp(now()).strftime('%b%dth%Y')
def getDefaultOutputFilename(name):
    (name, ext) = splitext(name)
    return '%s-%s.txt' % (name, nowstr())

def sendToPeer(peer, message, inline = False):
    message = '&w%s&N' % message
    if inline:
        peer.sendln(message)
    else:
        enqueueHeartbeatTask(peer.sendln, message)

def getArchiveIndex(filename):
    (name, ext) = splitext(filename)
    if ext.lower() == '.zip':
        return ZipArchiveIndex(filename)

    raise TypeError(ext)

def LoadWorld(index):
    w = World(index)
    w.loadWorld(cascade = True)
    return w

# Interface.
def parseOptions(args):
    parser = OptionParser()
    parser.add_option('-o', '--output')
    parser.add_option('-g', '--debug', action = 'store_true')
    parser.add_option('-u', '--local-pickle', '--unpickle')
    parser.add_option('-p', '--pickle')

    parser.add_option('--no-info', action = 'store_false', default = True,
                      dest = 'info')
    parser.add_option('--no-commands', action = 'store_false', default = True,
                      dest = 'commands')
    parser.add_option('--no-rooms', action = 'store_false', default = True,
                      dest = 'rooms')
    parser.add_option('--no-items', action = 'store_false', default = True,
                      dest = 'items')
    parser.add_option('--no-mobiles', action = 'store_false', default = True,
                      dest = 'mobiles')

    return parseOptionsOverSystem(parser, args)

@parameters(1)
def doCompareWorld(peer, *options):
    (options, args) = parseOptions(options)
    if options.pickle:
        # Actually, just load the world and then pickle it.
        def convertLocalWorld(pickleFile):
            e = Elapsed()
            w = LoadWorld(LocalIndex('world'))
            sendToPeer(peer, 'Load Time: %s\r\nPickling...' % e, inline = True)
            nth(dumpWorldPickle, (w, pickleFile))

        def dumpWorldPickle(w, pickleFile):
            e = Elapsed()
            pickle(w, open(pickleFile, 'w'))
            sendToPeer(peer, 'Time: %s' % e)

        sendToPeer(peer, 'Loading local world...', inline = True)
        enqueueHeartbeatTask(convertLocalWorld, options.pickle)
        return

    name = args[0] # XXX assertion

    filename = getCheckedPlayerFilename(peer, name, PERSONAL_PLRFILE_DIR)
    playerArchive = getArchiveIndex(filename)
    outputfile = getCheckedPlayerFilename(peer, options.output or getDefaultOutputFilename(name),
                                          PERSONAL_PLRFILE_DIR)

    def comparatorThread(this):
        try:
            e = Elapsed()
            that = LoadWorld(playerArchive)
            sendToPeer(peer, 'Load Time: %s\r\nComparing...' % e)

            e = Elapsed()
            result = compareWorlds(this, that,
                                   info = options.info,
                                   commands = options.commands,
                                   rooms = options.rooms,
                                   items = options.items,
                                   mobiles = options.mobiles)

            # The comparison result and analysis is bundled together as text,
            # so the result is expected to be writable to file, just so.
            fl = open(outputfile, 'wt')
            fl.write(result)
            fl.close()

            sendToPeer(peer, ('Analysis Complete: %s\r\n' +
                              'Time: %s') % (basename(outputfile), e))

            def confirmView(peer, line):
                line = line.strip().lower()
                if line in ['y', 'ye', 'yes']:
                    peer.page_string(open(outputfile).read())

            enqueueHeartbeatTask(PromptFor, peer, confirmView, message = 'View? ')

        except:
            enqueueHeartbeatTask(HandleCommandError, peer, exc = exc_info())

    # The advantage of loading from pickle is (not speed, curiously, but) to
    # do so in another thread and not block the mud.
    sendToPeer(peer, 'Loading local world...', inline = True)
    if options.local_pickle:
        # Loading a pickle can be done in another thread, because it's not
        # really the active world (even though it's in the place of it).
        def loadPickleAndStartComparison(pickleFile):
            try:
                e = Elapsed()
                this = unpickle(open(pickleFile))

                sendToPeer(peer, 'Load Time: %s\r\nLoading %s...' % (e, basename(filename)))
                comparatorThread(this)
            except:
                enqueueHeartbeatTask(HandleCommandError, peer, exc = exc_info())

        nth(loadPickleAndStartComparison, (options.local_pickle,))
    else:
        def startComparison():
            # This should be done within the heartbeat to synchronize with active db ops.
            try:
                e = Elapsed()
                this = LoadWorld(LocalIndex('world'))

                sendToPeer(peer, 'Load Time: %s\r\nLoading %s...' % (e, basename(filename)), inline = True)
                nth(comparatorThread, (this,))
            except:
                HandleCommandError(peer)

        enqueueHeartbeatTask(startComparison)

# Algorithm:
# todo: proper zcommands and exits comparison.
# todo: analysis resulting in stuph:document format for olc web deployment.
ZONE_ATTRIBUTES = ('continent', 'flags', 'name', 'resetModeName', 'top')
ROOM_ATTRIBUTES = ('descr', 'flags', 'name', 'sectorName')
ITEM_ATTRIBUTES = ('actiondescr', 'antiflags', 'cost', 'extraflags', 'longdescr',
                   'minlevel', 'name', 'shortdescr', 'type', 'wearflags', 'weight',
                   'value1', 'value2', 'value3', 'value4')
MOBILE_ATTRIBUTES = ('affbits', 'alignment', 'armorclass', 'damnodice',
                     'damroll', 'damsizedice', 'default_pos', 'descr',
                     'experience', 'gender', 'gold', 'hitroll', 'level',
                     'longdescr', 'max_hit', 'max_mana', 'max_move',
                     'mobflags', 'name', 'position', 'shortdescr')

def compareWorlds(this, that, info = True, commands = True, rooms = True, items = True, mobiles = True):
    def compareZones(these, those):
        for (vnum, zone) in these.items():
            try: z = those[vnum]
            except KeyError:
                yield [['zone', vnum], 'removed']
            else:
                if info:
                    for attribute in ZONE_ATTRIBUTES:
                        value = getattr(z, attribute)
                        if value != getattr(zone, attribute):
                            yield [['zone', vnum], [attribute, value]]

                if commands:
                    # yield filterNone(compareCommands(zone.zcommands, z.zcommands))
                    yield filterNone(compareObjects(getattr(zone, 'commands', []),
                                                    getattr(z, 'commands', [])))

                if rooms:
                    yield filterNone(compareRooms(zone.roomMap, z.roomMap))
                if items:
                    yield filterNone(compareItems(zone.itemMap, z.itemMap))
                if mobiles:
                    yield filterNone(compareMobiles(zone.mobileMap, z.mobileMap))

        for (vnum, zone) in those.items():
            if vnum not in these:
                yield [['zone', vnum], 'added', getFullZoneData(zone)]

    def compareCommands(these, those):
        return compareObjects([(c.arg1, c.arg2, c.arg3, c.conditional) for c in these],
                              [(c.arg1, c.arg2, c.arg3, c.conditional) for c in those])

    def compareRooms(these, those):
        for (vnum, room) in these.items():
            try: r = those[vnum]
            except KeyError:
                yield [['room', vnum], 'removed']
            else:
                for attribute in ROOM_ATTRIBUTES:
                    value = getattr(r, attribute)
                    if value != getattr(room, attribute):
                        yield [['room', vnum], [attribute, value]]

                # Now, analyze .exitdirs

        for (vnum, room) in those.items():
            if vnum not in these:
                yield [['room', vnum], 'added', getFullRoomData(room)]

    def compareItems(these, those):
        for (vnum, item) in these.items():
            try: i = those[vnum]
            except KeyError:
                yield [['item', vnum], 'removed']
            else:
                for attribute in ITEM_ATTRIBUTES:
                    value = getattr(i, attribute)
                    if value != getattr(item, attribute):
                        yield [['item', vnum], [attribute, value]]

        for (vnum, item) in those.items():
            if vnum not in these:
                yield [['item', vnum], 'added', getFullItemData(item)]

    def compareMobiles(these, those):
        for (vnum, mob) in these.items():
            try: m = those[vnum]
            except KeyError:
                yield [['mobile', vnum], 'removed']
            else:
                for attribute in MOBILE_ATTRIBUTES:
                    value = getattr(m, attribute)
                    if value != getattr(mob, attribute):
                        yield [['mobile', vnum], [attribute, value]]

        for (vnum, mob) in those.items():
            if vnum not in these:
                yield [['mobile', vnum], 'added', getFullMobileData(mob)]

    # Resulting output format (yaml):
    from yaml import dump
    return dump(filterNone(compareZones(this.zoneMap, that.zoneMap)))

# For now, just return a summary for reference.
def getFullZoneData(zone):
    return repr(zone)
def getFullRoomData(room):
    return repr(room)
def getFullItemData(item):
    return repr(item)
def getFullMobileData(mob):
    return repr(mob)

# Generic Routines.
def filterNone(sequence):
    return [_f for _f in sequence if _f]

def compareObjects(this, that):
    from yaml import dump
    from os import popen, unlink
    from os.path import exists, join as joinpath
    from random import choice

    def getRandomFilename(tmppath = '/tmp', alphabet = 'abcdef0123'):
        while True:
            filename = ''.join(choice(alphabet) for x in range(10))
            filename = joinpath(tmppath, filename)
            if not exists(filename):
                return filename

    fn1 = getRandomFilename()
    fn2 = getRandomFilename()
    fn3 = getRandomFilename()

    try:
        fl = open(fn1, 'w')
        fl.write(dump(this))
        fl.close()

        fl = open(fn2, 'w')
        fl.write(dump(that))
        fl.close()

        popen('diff %r %r > %r' % (fn1, fn2, fn3))
        return open(fn3).read()

    finally:
        unlink(fn1)
        unlink(fn2)
        unlink(fn3)



# Generic In-MUD XMLRPC Server.
# --
# Designed to be hooked into by various components.
# 
#
# todo: Access-Control-Max-Age
from xmlrpc.server import SimpleXMLRPCServer as XMLRPC
from xmlrpc.server import SimpleXMLRPCRequestHandler as XMLRPCRequestHandler
# from thread import start_new_thread as nth
from urllib.parse import urlparse
from errno import EINTR
from select import error as select_error
from threading import Thread

import xmlrpc.client
import sys
import json

from stuphos import enqueueHeartbeatTask, logException
from stuphos.etc import nth, reraiseSystemException, getSystemExceptionString, isYesValue
from stuphos.etc.tools.timing import getCurrentTime as now

ALLOWED_PORT_ORIGINS = [#8000,
                        2180, None, 443]

# CSRF Access Control.
def getOriginAccess(request, origin):
    # Todo: Obviously, this is seriously unfinished:
    # This should come from an access table defined via mud-control.
    if origin and urlparse(origin).port in ALLOWED_PORT_ORIGINS:
        # note: this is no longer good enough for firefox/chrome
        # It must match the value passed for the Origin: header by client.
        #return origin # extracted during do_POST
        return origin
        return '*'

def redirectStderr(function, *args, **kwd):
    import sys
    try: __stderr__ = sys.__stderr__
    except AttributeError:
        return function(*args, **kwd)
    else:
        stderr = sys.stderr
        sys.stderr = __stderr__
        try: return function(*args, **kwd)
        finally:
            sys.stderr = stderr

class Deferrable:
    # For GETs (handled but undispatched).
    deferred = False

    billableIdentity = None

    def handleAccessControl(self):
        # https://developer.mozilla.org/En/HTTP_Access_Control
        allowOrigin = getOriginAccess(self, getattr(self, 'origin', None))
        if allowOrigin:
            #print 'Sending header', allowOrigin
            self.send_header('Access-Control-Allow-Origin', allowOrigin)

    # Deferred Methods Branch:
    def finish_this_request(self, response, other_headers = None, status = 200):
        redirectStderr(self.send_response, status)

        if response is not None:
            responseEncoded = response.encode('utf-8')

            self.send_header("Content-type", "text/xml")
            self.send_header("Content-length", str(len(responseEncoded)))

        self.handleAccessControl()
        self.send_header('Access-Control-Allow-Credentials', 'true')

        if other_headers:
            for (name, value) in other_headers.items():
                #print 'Sending other header', name, value
                self.send_header(name, value)

        self.end_headers()

        if response is not None:
            # print(response)
            # if response:
            #     print(response)
            #     debugOn()

            # bug in debugger?  just putting debugOn here means
            # that any subsequent post-preflight (options) request
            # will be somehow corrupted by the debugger and cause
            # it to return effectively xmlrpclib.dumps('[]')
            # instead of the proper (offending sample) response.

            self.wfile.write(responseEncoded)

        # shut down the connection
        #   File "/usr/lib/python2.7/socket.py", line 307, in flush
        #     self._sock.sendall(view[write_offset:write_offset+buffer_size])
        # AttributeError: 'NoneType' object has no attribute 'sendall'
        try: self.wfile.flush()
        except AttributeError as e:
            if str(e) != "'NoneType' object has no attribute 'sendall'":
                raise e

        self.connection.shutdown(1)

        identity = getattr(self, 'billableIdentity', None)
        if identity is not None:
            self.chargeUserContent(identity, response)


    def chargeUserContent(self, identity, response):
        pass


    def finish(self):
        if not self.deferred:
            if not self.wfile.closed:
                self.wfile.flush()
            self.wfile.close()
            self.rfile.close()

            # Carried over from server.close_request.
            self.request.close()

    # Deferment Request Methods:
    def defer(self):
        # Go into deferred mode, so that this request handler finishes
        # its construction phase, but the connection isn't closed yet.
        self.deferred = True
        raise HostRpc.DeferredMethodSignal(self)


class HostRpc(XMLRPC):
    # Passed to server on construction.
    LOG_REQUESTS = True

    class DeferredMethodSignal(Exception):
        # Signals a method that will be finished by another thread!
        def __init__(self, request):
            self.request = request

    class RequestHandler(Deferrable, XMLRPCRequestHandler):
        def setup(self):
            # XXX :network:
            self.request.settimeout(30)
            XMLRPCRequestHandler.setup(self)

        # def log_message(self, format, request_line, status = '-', *args, **kwd):
            # Todo: log to sys/mudlog.
            # return XMLRPCRequestHandler.log_message(self, *args, **kwd)
            # debugOn()

        def log_request(self, code='-', size='-'):
            """Log an accepted request.

            This is called by send_response().

            """
            from http import HTTPStatus

            if isinstance(code, HTTPStatus):
                code = code.value
            # self.log_message('"%s" %s %s',
            #                  self.requestline, str(code), str(size))

            request_line = self.requestline
            status = code
            args = ()

            # Reformat logging of RPC.

            # if httpMethod is not 'OPTIONS':
            rpcMethod = getattr(self, 'method', '-')

            # self.logRpcMessage(rpcMethod, getattr(self, 'params', []))

            if self.ignoreRpcMethod(rpcMethod):
                return

            httpMethod = request_line.split(' ', 1)[0]
            args = f'RPC {httpMethod} {rpcMethod} {", ".join(map(str, args))}'

            # debugOn()
            # Todo: logOperation
            from sys import stderr
            stderr.write(f'{self.address_string()} - - [{self.log_date_time_string()}] {repr(args)} {status} -\n')

        @runtime.available(runtime.System.Journal)
        def logRpcMessage(log, self, method, params):
            if method == 'session.command' and params and \
                params[0] in ['do-interpret-script',
                              'do-interpret-agent']:

                content = json.dumps(dict(code = params[1])) # Todo: params..?

                log += dict(source = f'rpc:{self.address_string()}',
                            type = 'webrequest-rpc/command/json',
                            timestamp = now(),
                            content = content)

        def ignoreRpcMethod(self, name):
            return name in ['session.messages', 'session.command']

        ##    def setup(self):
        ##        # StreamRequestHandler.setup(self)
        ##        XMLRPCRequestHandler.setup(self)
        ##
        ##        class socketWrapper(object):
        ##            def __init__(self, realsocket):
        ##                self.__dict__['__realsocket'] = realsocket
        ##            def __setattr__(self, name, value):
        ##                return setattr(self.__dict__['__realsocket'], name, value)
        ##            def __getattr__(self, name):
        ##                try: return getattr(self.__dict__['__realsocket'], name)
        ##                finally:
        ##                    if name in ['close', 'shutdown']:
        ##                        print 'tracing...'
        ##
        ##                        from pdb import set_trace
        ##                        set_trace()
        ##
        ##        self.request = socketWrapper(self.request)
        ##        self.connection = socketWrapper(self.connection)

        # What about adding 'internal' to this??
        rpc_paths = XMLRPCRequestHandler.rpc_paths + ('/rpc', '') # '/stuph/rpc'

        ##    def setup(self):
        ##        self.rpc_paths = self.rpc_paths + (str(self.request.path),)
        ##        XMLRPCRequestHandler.setup(self)

        import re
        METHOD_MATCH = r'[a-zA-Z_\.\ ]+' # r'*?' # Non-greedy.
        UNSUPPORTED_METHOD_MSG = re.compile(r'method "' + METHOD_MATCH + r'" is not supported')

        # XXX This method is no longer needed since the override is rewritten in the server class:
        # (However, log message relies on its functionality)

        def _dispatch(self, method, params):
            # Intercept method dispatch to see if it was recognized.
            # Also, prepend this handler instance to the arglist.

            # For log_message when this request finishes.
            self.method = method
            self.params = params

            try: result = self.server._dispatch(self, method, params)
            except HostRpc.DeferredMethodSignal:
                # Todo: filter out logging for certain methods.
                # self.log_message('RPC: %r' % str(method))
                raise

            except Exception as e:
                if self.UNSUPPORTED_METHOD_MSG.match(str(e)):
                    # Log unsupported method call.
                    self.log_message('UNKNOWN: %r {%s}', str(method), ', '.join(map(repr, params)))
                else:
                    # self.log_message('ERROR:\n%s', getSystemExceptionString())
                    logException(traceback = True)

                reraiseSystemException()

            # Log supported method call.
            # Todo: filter out logging for certain methods.
            # self.log_message('RPC: %r', str(method))
            return result

        def do_OPTIONS(self):
            # Handle CORS Preflight check.
            #print inspect(self.headers)

            ##    dict                           (dict)               {'origin': 'http://localhost:8000', 'accept-language': 'en-U
            ##    encodingheader                 (NoneType)           None
            ##    fp                             (socket._fileobject) <socket._fileobject object at 0x04EF38B0>
            ##    headers                        (list)               ['Host: localhost:2172\r\n', 'User-Agent: Mozilla/5.0 (Windo
            ##    maintype                       (str)                'text'
            ##    plist                          (list)               []
            ##    plisttext                      (str)                ''
            ##    seekable                       (int)                0
            ##    startofbody                    (NoneType)           None
            ##    startofheaders                 (NoneType)           None
            ##    status                         (str)                ''
            ##    subtype                        (str)                'plain'
            ##    type                           (str)                'text/plain'
            ##    typeheader                     (NoneType)           None
            ##    unixfrom                       (str)                ''

            # print nls(self.headers.headers)
            # print self.headers # MIME Message object.

            ##    Host: localhost:2172
            ##    User-Agent: Mozilla/5.0 (Windows NT 6.1; rv:30.0) Gecko/20100101 Firefox/30.0
            ##    Accept: text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8
            ##    Accept-Language: en-US,en;q=0.5
            ##    Accept-Encoding: gzip, deflate
            ##    Origin: http://localhost:8000
            ##    Access-Control-Request-Method: POST
            ##    Access-Control-Request-Headers: content-type
            ##    Connection: keep-alive
            ##    Pragma: no-cache
            ##    Cache-Control: no-cache

            self.origin = self.headers.get('Origin')
            self.finish_this_request('', {'Access-Control-Allow-Methods': 'GET, POST, OPTIONS', # Should this include OPTIONS?
                                          # Q: Are OPTIONS requests sending headers??
                                          'Access-Control-Allow-Headers': 'Content-Type'})

            # print 'finished OPTIONS'


        # Overwrite original stdlib method for deferred methods:
        def do_POST(self):
            """Handles the HTTP POST request.

            Attempts to interpret all HTTP POST requests as XML-RPC calls,
            which are forwarded to the server's _dispatch method for handling.

            Some methods may be deferred for another thread to finish.
            """

            # Check that the path is legal
            if not self.is_rpc_path_valid():
                self.report_404()
                return

            self.deferred = False

            try:
                # :set-user: :authentication: note: this only returns a dictionary...
                self.user = self.authenticate(self.headers.get('Authorization'))


                # from python3.9:
                # Get arguments by reading body of request.
                # We read this in chunks to avoid straining
                # socket.read(); around the 10 or 15Mb mark, some platforms
                # begin to have problems (bug #792570).
                max_chunk_size = 10*1024*1024
                size_remaining = int(self.headers["content-length"])
                L = []

                # self.socket.setblocking(0)
                # self.socket.settimeout(10)

                while size_remaining:
                    chunk_size = min(size_remaining, max_chunk_size)
                    # if chunk_size == 2: # XXX :skip: HACK
                    #     print('setting blocking to 0')
                    #     self.socket.setblocking(0)

                    chunk = self.rfile.read(chunk_size)
                    if not chunk:
                        break
                    L.append(chunk)
                    size_remaining -= len(L[-1])

                    # print(f'[xmlrpc] {size_remaining}')

                # self.socket.setblocking(1)
                # self.socket.settimeout(None)

                # if size_remaining:
                #     print(f'[xmlrpc] incomplete read')

                data = b''.join(L)

                data = self.decode_request_content(data)
                if data is None:
                    # print('[xmlrpc] response has been sent')
                    return #response has been sent


                # debugOn()

                # max_chunk_size = 10*1024*1024
                # size_remaining = int(self.headers["content-length"])
                # L = []
                # while size_remaining:
                #     chunk_size = min(size_remaining, max_chunk_size)

                #     # XXX :skip: if unicode characters are included, the size_remaining will be wrong..?
                #     # Meaning we'll read more than is available and block (forever).
                #     L.append(self.rfile.read(chunk_size).decode('utf-8'))
                #     size_remaining -= len(L[-1])
                # data = ''.join(L)


                # This qualifies as a "preflight" because of POST method and content-type.
                self.origin = self.headers.get('Origin')

                contentType = self.headers.get('Content-Type', 'application/xml')
                response = self.server.handleContentType(contentType, self, data)

                if response is None:
                    # debugOn()

                    # In previous versions of SimpleXMLRPCServer, _dispatch
                    # could be overridden in this class, instead of in
                    # SimpleXMLRPCDispatcher. To maintain backwards compatibility,
                    # check to see if a subclass implements _dispatch and dispatch
                    # using that method if present.
                    response = self.server._marshaled_dispatch(
                            data, getattr(self, '_dispatch', None)
                        )

                # print(f'[xmlrpc] response: {response}')

            except HostRpc.DeferredMethodSignal:
                # Done (for now).
                ##    print 'tracing...'
                ##
                ##    from pdb import set_trace
                ##    set_trace()

                return

            except:
                # This should only happen if the module is buggy
                # internal error, report as HTTP server error
                self.error_this_request(fault = False)
            else:
                self.finish_this_request(response)

                # todo: debug
                #print 'finished POST'

        def authenticate(self, header):
            from django.contrib.auth import authenticate
            if header and header.startswith('Basic '):
                auth = header[6:]

                import base64
                auth = base64.decodestring(auth)
                (username, password) = auth.split(':')

                return authenticate(username = username,
                                    password = password)

        # Q: Write do_GET for preflight checks?!

        def error_this_request(self, fault = True):
            # Q: Log error?
            e = getSystemExceptionString()
            # debugOn()
            self.log_message('SERVER-ERROR:\n%s', e)

            if fault:
                response = self.server._dump_method_response(xmlrpc.client.Fault(1, e))
            else:
                response = None

            self.finish_this_request(response, status = 500)
            self.end_headers()

        def complete(self, response):
            self.deferred = False

            # debugOn()
            try: response = self.server._dump_method_response(response)
            except: self.error_this_request()
            else: self.finish_this_request(response)

            try: self.finish()
            except AttributeError as e:
                if str(e) != "'NoneType' object has no attribute 'close'":
                    raise e


    def __init__(self, config, methods = None):
        XMLRPC.__init__(self, (config.hostname, config.port),
                        requestHandler = self.RequestHandler,
                        logRequests = self.LOG_REQUESTS)

        if config.certificate is not None:
            import ssl
            self.socket = ssl.wrap_socket (self.socket, certfile = config.certificate,
                                           keyfile = config.keyfile,
                                           server_side = True)

        if methods is not None:
            self.funcs = methods

        self.register_introspection_functions()
        self.register_multicall_functions()

    def _marshaled_dispatch(self, data, dispatch_method = None):
        # This override passes as much of the exception data as part
        # of the fault structure value.  It could also implement JSON!
        #
        # debugOn()

        try:
            # Todo: parse 'parsererror' element in data.
            params, method = xmlrpc.client.loads(data)

            # generate response
            if dispatch_method is not None:
                response = dispatch_method(method, params)
            else:
                response = self._dispatch(None, method, params)
            # wrap response in a singleton tuple
            response = (response,)
            response = xmlrpc.client.dumps(response, methodresponse=1,
                                       allow_none=self.allow_none, encoding=self.encoding)

        except xmlrpc.client.Fault as fault:
            response = xmlrpc.client.dumps(fault, allow_none=self.allow_none,
                                       encoding=self.encoding)

        except xmlrpc.client.ResponseError as e:
            print(f'[xmlrpc$dispatch] {e}')
            print(data)

            raise

        except self.DeferredMethodSignal:
            raise

        except Exception as e:
            # report exception back to server -- X-exception, X-traceback ??
            info = getSystemExceptionString()

            if isinstance(e, ValueError) and isYesValue(configuration.XMLRPC.show_parse_source):
                info = f'{data.decode()}\n{info}'

            response = xmlrpc.client.dumps(
                xmlrpc.client.Fault(1, info),
                encoding=self.encoding,
                allow_none=self.allow_none,
                )

        # debugOn()
        return response

    contentTypeHandlers = {'application/json': None}

    def handleContentType(self, contentType, request, data):
        try: handler = self.contentTypeHandlers[contentType]
        except KeyError: pass
        else: return handler(request, data)

    def handleMethod(self, request, method, *params):
        return self._dispatch(request, method, params)

    def _handleMethodDjango(self, request, method, *params):
        return self._dispatch(request, method, params)

    def handleMethodDjango(self, request, method, *params):
        'Handle a request coming from Django WSGI application.  May defer.'

        return self._handleMethodDjango(self.DjangoRequestWrapper(request), method, params)

    class DjangoRequestWrapper(Deferrable):
        # Powerful bridge between XMLRPC request handler and django request object.
        def __init__(self, django):
            self.django = django

        def address_string(self):
            return ''

        # .wfile.closed, .wfile.flush(), .wfile.close(), .rfile.close()
        # .request.close(), .connection.shutdown(1), .wfile.write(response)
        # .end_headers(), .send_header(name, value), .origin, .send_response()

        def complete(self, msgs):
            self.finish_this_request(msgs)
            self.finish()

        @property
        def META(self):
            # REMOTE_ADDR
            return self.django.META

        @property
        def user(self):
            return self.django.user

        @property
        def server(self):
            return self.django.server


    def close_request(self, request):
        # Request handler class differentiates between closable socket modes.
        pass

    def shutdown_request(self, request):
        # XXX request is the socket._socketobject
        ##    if not request.deferred:
        ##        XMLRPC.shutdown_request(self, request)
        pass

    # Overridden to expose the request object.
    def _dispatch(self, request, method, params):
        """Dispatches the XML-RPC method.

        XML-RPC calls are forwarded to a registered function that
        matches the called XML-RPC method name. If no such function
        exists then the call is forwarded to the registered instance,
        if available.

        If the registered instance has a _dispatch method then that
        method will be called with the name of the XML-RPC method and
        its parameters as a tuple
        e.g. instance._dispatch('add',(2,3))

        If the registered instance does not have a _dispatch method
        then the instance will be searched to find a matching method
        and, if found, will be called.

        Methods beginning with an '_' are considered private and will
        not be called.

        All methods are passed the server instance and the request
        handler object instance as the first parameters.
        """

        # print(f'[xmlrpc.host.dispatch] {method} {", ".join(map(repr, params))}')
        # if method == 'session.command' and params[1] == 'do-interpret-script':
        #     debugOn()

        func = None
        try:
            # check to see if a matching function has been registered
            func = self.funcs[method]
        except KeyError:
            if self.instance is not None:
                # check for a _dispatch method
                if hasattr(self.instance, '_dispatch'):
                    return self.instance._dispatch(method, params)
                else:
                    # call instance method directly
                    try:
                        func = resolve_dotted_attribute(
                            self.instance,
                            method,
                            self.allow_dotted_names
                            )
                    except AttributeError:
                        pass

        if func is not None:
            # Passes this server instance and the request handler object instance.
            # print(f'[xmlrpc] func: {func}')
            return func(self, request, *params)
        else:
            raise Exception('method "%s" is not supported' % method)

    def _dump_method_response(self, response):
        response = (response,)
        response = xmlrpc.client.dumps(response, methodresponse=1,
                                   allow_none=self.allow_none,
                                   encoding=self.encoding)
        return response

    # Compatibility Overrides.
    def system_listMethods(self, host, request):
        return XMLRPC.system_listMethods(self, )
    def system_methodSignature(self, host, request, method_name):
        return XMLRPC.system_methodHelp(self, method_name)
    def system_methodHelp(self, host, request, method_name):
        return XMLRPC.system_methodHelp(self, method_name)


    # why?
    # http.server:419
    # def handle_one_request(self):
    #     except socket.timeout as e:
    #         #a read or a write timed out.  Discard this connection
    #         self.log_error("Request timed out: %r", e)


    def handle_request(self):
        while True:
            # import pdb; pdb.set_trace()
            try: return XMLRPC.handle_request(self)
            except select_error as e:
                if e.args[0] is EINTR:
                    # A process signal was sent, ignore and continue.
                    continue

                (etype, value, tb) = sys.exc_info()
                raise etype(value).with_traceback(tb)

    def _handle_request_noblock(self):
        'Override method to explicitly handle socket errors.'
        import socket
        try:
            request, client_address = self.get_request()
        except socket.error as e:
            if not str(e).startswith('[SSL: SSLV3_ALERT_CERTIFICATE_UNKNOWN]'):
                print(f'[_handle_request_noblock]: {e}')

            return

        if self.verify_request(request, client_address):
            try:
                self.process_request(request, client_address)
            except:
                self.handle_error(request, client_address)
                self.shutdown_request(request)
        else:
            self.shutdown_request(request)

    def handle_error(self, request, client_address):
        from stuphos.system.api import syslog
        from stuphos import logException
        syslog('Exception happened during processing of request from %s' % (client_address,))
        logException(traceback = True)

    # Todo: expound on the result-bearing functionality.
    def register_heartbeat_function(self, function, result, *args, **kwd):
        XMLRPC.register_function(self, self._get_heartbeat_dispatcher(function, result),
                                 *args, **kwd)

    def _get_heartbeat_dispatcher(self, function, result):
        def heartbeat_dispatcher(*args, **kwd):
            enqueueHeartbeatTask(function, *args, **kwd)
            # XXX :skip: This is wrong, isn't it?  It should be called from the heartbeat?
            # Or, maybe what I want is for this function to block, waiting for the
            # heartbeat.  In which case, we'll need to program some synchronicity
            # primitives here (like Queue).
            if callable(result):
                return result()

        return heartbeat_dispatcher

    def unregister_function(self, name):
        try: del self.funcs[name]
        except KeyError:
            pass

    # Server-Control.
    def is_running(self):
        try: return self.__running
        except AttributeError:
            return False

    def start(self):
        if not self.is_running():
            self.__running = True
            Thread(target = self.serve_cooperatively, daemon = True).start()
            # nth(self.serve_cooperatively)

    def stop(self):
        try: del self.__running
        except AttributeError:
            pass

        self.server_close()

    def serve_cooperatively(self):
        try:
            while self.is_running():
                self.handle_request()

        finally:
            # This might not be proper form, but it's a shortcut to cleanup.
            self.stop()


    # def server_bind(self):
    #     self.socket.setblocking(0) # XXX :skip: What I want is nonblocking peers.
    #     return XMLRPC.server_bind(self)


    ##    def __registry_delete__(self):
    ##        self.stop()


def installMarshaller():
    from stuphos.kernel import StringValue, baseStringClass

    class AgentMarshaller(xmlrpc.client.Marshaller):
        dispatch = xmlrpc.client.Marshaller.dispatch.copy()
        dispatch[StringValue] = dispatch[str]
        dispatch[baseStringClass] = dispatch[str]

        @property
        def allow_none(self):
            return True

        @allow_none.setter
        def allow_none(self, value):
            'Compat with initializer.'


    # This is the interface used by dumps.
    xmlrpc.client.FastMarshaller = AgentMarshaller



from stuphlib.binary import Record, BuildRecordClass, ReadFile

class RentedItemData(Record, metaclass=BuildRecordClass):
    # obj_file_elem
    class Types:
        class ItemAffectedType(Record, metaclass=BuildRecordClass):
            # obj_affected_type
            Fields = [('location', Record.Byte),
                      ('modifier', Record.SByte)]

        NAME_LENGTH        = 128
        DESCR_LENGTH       = 256
        SHORT_DESCR_LENGTH = 128

        NUM_VALUES         = 4
        MAX_OBJ_AFFECT     = 6

        ObjVnum            = Record.Int

        NameArray          = Record.CString(NAME_LENGTH)
        DescrArray         = Record.CString(DESCR_LENGTH)
        ShortDescrArray    = Record.CString(SHORT_DESCR_LENGTH)

        ValueArray         = Record.Int       [NUM_VALUES        ]

        AffectedArray      = Record.Array(ItemAffectedType, MAX_OBJ_AFFECT)

    Fields = [('item_number'      , Types.ObjVnum          ),
              ('name'             , Types.NameArray        ),
              ('description'      , Types.DescrArray       ),
              ('short_description', Types.ShortDescrArray  ),
              ('locate'           , Record.ShInt           ),
              ('value'            , Types.ValueArray       ),
              ('item_type'        , Record.Int             ),
              ('extra_flags'      , Record.Int             ),
              ('anti_flags'       , Record.Int             ),
              ('wear_flags'       , Record.Int             ),
              ('weight'           , Record.Int             ),
              ('timer'            , Record.Int             ),
              ('bitvector'        , Record.Long            ),
              ('affected'         , Types.AffectedArray    )]

class RentFileInfo(Record, metaclass=BuildRecordClass):
    # rent_info
    Fields = [('time',              Record.Int),
              ('rentcode',          Record.Int),
              ('net_cost_per_diem', Record.Int),
              ('gold',              Record.Int),
              ('account',           Record.Int),
              ('nitems',            Record.Int),
              ('spare0',            Record.Int),
              ('spare1',            Record.Int),
              ('spare2',            Record.Int),
              ('spare3',            Record.Int),
              ('spare4',            Record.Int),
              ('spare5',            Record.Int),
              ('spare6',            Record.Int),
              ('spare7',            Record.Int)]

def LoadRentFile(fp):
    # Crash_load (of sorts -- interprets locate field)
    rent = ReadFile(fp, RentFileInfo, 1)
    model = dict(rent      = list(rent)[0],
                 inventory = []  ,
                 equipment = dict())

    for item in ReadFile(fp, RentedItemData):
        # XXX Put all in inventory
        model['inventory'].append(item)

    return model

# Inspection Front End.
def main(argv = None):
    from optparse import OptionParser
    from os.path import basename, splitext
    parser = OptionParser()

    (options, args) = parser.parse_args(argv)
    for filename in args:
        if basename(filename).lower() == 'plrobjs':
            # Inspect directory (using middle initials)
            pass

        elif splitext(filename)[1].lower() == '.objs':
            model = LoadRentFile(open(filename))
            print(model['rent'])
            print('\n'.join(model['inventory']))

            for (where, item) in model['equipment'].items():
                print('%s: %r' % (where, item))

if __name__ == '__main__':
    main()



# XDebug front-end.
from socketserver import ThreadingTCPServer, StreamRequestHandler
from xml.dom.minidom import parseString
from io import StringIO as NewBuffer
from _thread import start_new_thread as nth

from stuphos.runtime.facilities import Facility
from stuphos import enqueueHeartbeatTask, executeInline
from .wizard import parameters

from stuphos.system.api import mudlog
import imp


DBGP_COMMANDS = \
'''&yDGBp Commands&N
&r=============&N
breakpoint_list
breakpoint_{get,remove,update} -d <name>
  update: -s {enabled,disabled}
          -o {>=,==,%}
          -n <line-number>
          -h <hit-value>

breakpoint_set
  -t {line,conditional,call,return,exception,watch}
  -n <line-number>
  -f <file-name>
  -m <function-name>
  -a <class-name>
  -x <exception-name>
  -s {enabled,disabled}
  -o {>=,==,%}
  -h <hit-value>
  -r <integer>
  -- <conditional-expression>

context_names
context_get
  -c <context # (integer)>
  -d <depth (integer)>

eval
  -p <page (integer)>
  -- <expression>

feature_{get,set}
  -v <value>
  -n {breakpoint_{languages,types},data_encoding,encoding,
      language_{name,supports_threads,version},
      max_{children,data,depth}, protocol_version,
      supported_encodings, supports_{async,postmortem},
      show_hidden}

typemap_get
property_{get,set,value}
  -d <depth (integer)>
  -c <context # (integer)>
  -p <page (integer)>
  -m <max data (integer)>
  -n <name>

  set: -t {bool,int,float,string}

source
  -f <file-name>
  -b <begin (integer)>
  -e <end (integer)>

stack_depth
stack_get
  -d <depth (integer)>

status
stderr
stdout
  -c {0,1}

run
step_{into,out,over}
stop
detach
xcmd_{profiler_name_get,get_executable_lines}
'''

class XDebugServer(Facility, ThreadingTCPServer):
    NAME = 'XDebug::Server'

    BIND_HOST = 'localhost'
    BIND_PORT = 9000

    class Manager(Facility.Manager):
        MINIMUM_LEVEL = Facility.Manager.IMPLEMENTOR
        VERB_NAME = 'xdebug*'

        def do_bind(self, peer, cmd, args):
            s = self.facility.get()
            if s is not None:
                if len(args) == 1:
                    key = args[0]
                    session = s.getSession(key)
                    if session is None:
                        print('Unknown session: %r' % key, file=peer)
                    else:
                        if session.isListening(peer):
                            print('Already listening to %s.' % key, file=peer)
                        else:
                            session.bind(session.PeerListener(peer))
                            print('Binded to %s.' % key, file=peer)

        def do_reload(self, peer, cmd, args):
            self.facility.destroy()
            module = imp.reload(__import__(__name__, globals(), locals(), ['']))
            self.facility.get(create = True)

            print(module, file=peer)

        def do_commands(self, peer, cmd, args):
            peer.page_string(DBGP_COMMANDS)

    @classmethod
    def create(self):
        return self((self.BIND_HOST, self.BIND_PORT),
                    self.XDebugDBGpRequestHandler)

    def __init__(self, binding, rqh):
        self.sessions = {}
        ThreadingTCPServer.__init__(self, binding, rqh)
        nth(self.serve, ())

    def serve(self):
        self.__running = True
        while self.__running:
            self.handle_request()

    def server_stop(self):
        self.__running = False
        self.server_close()

    def __registry_delete__(self):
        self.server_stop()

    def getStatus(self):
        sessions = '\n  '.join('%s -- %s' % (key, s) for (key, s) in self.sessions.items())
        return '%s (%s)%s%s\n' % (self.NAME, self.__class__.__name__,
                                sessions and ':\n  ' or '', sessions)

    def log(self, message):
        mudlog('[%s] %s' % (self.NAME, message))

    # Session Management.
    def addSession(self, key, requestHandler):
        if key not in self.sessions:
            self.sessions[key] = self.Session(key, requestHandler)
            self.log('New Session: %r' % key)

    def removeSession(self, key):
        if key in self.sessions:
            del self.sessions[key]
            self.log('Session Complete: %r' % key)

    def getSession(self, key):
        return self.sessions.get(key)

    class Session:
        class PeerListener:
            def __init__(self, peer):
                self.peer = peer
            def __eq__(self, other):
                return self.peer == other

            def notify(self, msg):
                try: m = self.peer.xdebug_messages
                except AttributeError:
                    m = self.peer.xdebug_messages = []

                m.append(msg)
                self.peer.page_string(describeMessage(msg))

        def __init__(self, key, requestHandler):
            self.key = key
            self.requestHandler = requestHandler
            self.listeners = []

        def isListening(self, object):
            for o in self.listeners:
                if o == object:
                    return True

        def bind(self, object):
            self.listeners.append(object)

        def sendCommand(self, command, args):
            self.requestHandler.sendCommand(self.key, command, args)

        def handleResponse(self, message):
            def sendNotifications():
                for o in self.listeners:
                    o.notify(message)

            enqueueHeartbeatTask(sendNotifications)

        def __str__(self):
            return ', '.join('%s=%r' % (name, value) for (name, value) in \
                             ((name, getattr(self, name, None)) for name in \
                              XDebugServer.INIT_VALUES) if value is not None)

    INIT_VALUES = ('protocol_version', 'language', 'appid', 'fileuri')

    # DBGp protocol
    class XDebugDBGpRequestHandler(StreamRequestHandler):
        def handle(self):
            # Read first message to initialize the session.
            init = self.readMessage().getElementsByTagName('init')[0]
            idekey = str(init.getAttribute('idekey'))

            executeInline(self.server.addSession, idekey, self)
            session = self.server.getSession(idekey)
            for name in XDebugServer.INIT_VALUES:
                setattr(session, name, str(init.getAttribute(name)))

            # Read all following message responses and forward to facility management.
            for msg in self:
                session.handleResponse(msg)

            executeInline(self.server.removeSession, idekey)

        def sendCommand(self, key, command, args):
            payload = [command, '-i', key]
            if args: payload.append(args)
            payload = ' '.join(payload)
            payload += '\x00'
            self.wfile.write(payload)

        def __iter__(self):
            while True:
                msg = self.readMessage()
                if msg is None:
                    break

                yield msg

        def readMessage(self):
            msg = self.readXmlMessage(self.rfile)
            if msg is not None:
                return parseString(msg)

        # what about socket.error: errno 104 Connection reset by peer?
        def readXmlMessage(self, conn):
            size = self.readMessageSize(conn)
            if size is not None:
                msg = conn.read(size)
                conn.read(1) # null-byte
                return msg

        def readMessageSize(self, conn):
            buf = ''
            while True:
                c = conn.read(1)
                if c == '':
                    break

                if c == '\x00':
                    return int(buf)

                buf += c

def describeMessage(msg):
    docEl = msg.documentElement
    if docEl.tagName == 'response':
        attrs = '\r\n'.join('%s: %s' % (str(attr.name).capitalize(), str(attr.value)) \
                            for attr in list(docEl.attributes.values()) \
                                if attr.name != 'xmlns' and attr.prefix is None)

        nodes = '\r\n'.join(getNodeValue(node) for node in docEl.childNodes)

        return '%s\r\n%s\r\n' % (attrs, nodes)

    elif docEl.tagName == 'stream':
        encoding = docEl.getAttribute('encoding')
        return ('Encoding: %s\r\n' +
                'Type: %s\r\n' +
                '%s\r\n') % (encoding, docEl.getAttribute('type'),
                             getTextContent(docEl).decode(encoding))

    return msg.toprettyxml()

def getNodeValue(node):
    if node.tagName == 'property':
        type = node.getAttribute('type')
        size = node.getAttribute('size')
        address = node.getAttribute('address')

        # string
        encoding = node.getAttribute('encoding')

        # array
        numChildren = node.getAttribute('numchildren')
        page = node.getAttribute('page')
        pagesize = node.getAttribute('pagesize')

        # array, object
        children = node.getAttribute('children')

        # and: 'name'
        # type: bool, int, float, string, null, array, object, property
        # type: resource, null

        value = getTextContent(node)
        if encoding == 'base64':
            value = value.decode('base64')

        # todo: array types have child nodes...

        return '%s\r\n%s' % ('\r\n'.join('%s: %s' % (str(name).capitalize(), str(value)) for (name, value) in \
                             [_f for _f in [('type', type), ('size', size), ('address', address),
                               ('encoding', encoding), ('page', page), ('pageSize', pagesize),
                               ('children', children), ('numChildren', numChildren)] if _f]),
                             value)
                
    return node.toprettyxml(indent = '  ') 

def getTextContent(node):
    return ''.join(c.data for c in node.childNodes \
                   if c.nodeType in (c.TEXT_NODE, c.CDATA_SECTION_NODE))

# Install.
XDebugServer.manage()
# XDebugServer.get(create = True)

# Wizard Control.
def getServerSession(key):
    server = XDebugServer.get()
    if server is not None:
        return server.getSession(key)

@parameters(2, None)
def doXDebug(peer, key, command, *args):
    session = getServerSession(key)
    if session is not None:
        # Some command preprocessing:
        if command == 'eval':
            i = args.index('--')
            assert i >= 0
            code = ' '.join(args[i+1:])
            code = code.encode('base64')

            args = list(args[:i+1])
            args.append(code)

        # Dispatch the command to the session.
        session.sendCommand(command.replace('-', '_'), ' '.join(args))



from op.compute.emulation.application import VM as _VM

def enter(frame, procedure, *args, **kwd):
	'''
	components/language/application::
		def init$(source):
			a = $.components.language.cpp(source)
			i = $.components.language.x86.load(a)
			m = i.main.emulationCompiled.function()
			t = $.components.language.vm.enter(m)

			def execute():
				while true:
					t.run()

			.task = task(execute)
			.library = i

			'''

	task = _VM.Task()
	task.frameCall(procedure, *args, **kwd)
	vm += task

	return task



from sqlobject import *

class StuphZoneCommand(SQLObject):
	type            = StringCol(length = 1)
	ifFlag          = IntCol()

	arg1            = IntCol()
	arg2            = IntCol()
	arg3            = IntCol()

	stuphZoneRecord = ForeignKey('StuphZoneRecord')

class StuphZoneRecord(SQLObject):
	# StuphZoneRecord.byVnum(12)
	vnum      = IntCol(alternateID = True)

	name      = StringCol()

	top       = IntCol()
	flags     = IntCol()
	lifespan  = IntCol()
	resetMode = IntCol()
	continent = IntCol()

	# zone = StuphZoneRecord(**Record)
	# StuphZoneCommand(stuphZoneRecord = zone, **Command)

	commands  = MultipleJoin('StuphZoneCommand')

	# RelatedJoin??
	rooms     = MultipleJoin('StuphRoomRecord')
	items     = MultipleJoin('StuphItemRecord')



# Software Architecture
from contextlib import contextmanager
# from types import ClassType as newClassObject, ClassType
newClassObject = type
TypeType = type
ClassType = type

def classObject(name, *bases, **values):
	return newClassObject(name, bases, values)
def subClass(cls, *bases, **values):
	return classObject(cls.__name__, cls, *bases, **values)

def call(o, *args, **kwd):
	'call(classObject("myClass", baseClass, action = callable), value, keyword = value)'

	return o(*args, **kwd)


def classOf(o): return o.__class__
def nameOf(o): return o.__name__
def dictOf(o): return o.__dict__


class applyWith:
	def __init__(self, *args, **kwd):
		self.args = args
		self.kwd = kwd

	def __call__(self, object):
		return object(*self.args, **self.kwd)


class connective:
	# Coupling Abstraction.
	pass

class event:
	'Describes an action.'


class scopeObject(object): # todo: new-style class?
	def __getitem__(self, name):
		try: return getattr(self, name)
		except AttributeError:
			raise KeyError(name)

class namespace(dict):
	def __init__(self, **kwd):
		self.__dict__ = self
		self.update(kwd)

# class extendMap:
# 	def __init__(self, map):
# 		self.map = map
# 	def __call__(self, **kwd):
# 		self.map.update(kwd)
# 		return self


class action:
	# Curry.

	def __init__(self, callable, *args, **kwd):
		self.callable = callable
		self.args = args
		self.kwd = kwd

	def __call__(self, *args, **kwd):
		return self.callable(*(self.args + args),
							 **dict(self.kwd, **kwd))

deferred = callable = action

class operator(property):
	'''
	class o:
		@operator
		class p(extension):
			pass

	class extendsO(o):
		@operator
		class p(o.p):
			pass

			'''

	# Note that if you subclass or reference an operator defined on a base class,
	# this descriptor will return that base class, so you must wrap it in another
	# operator property:

	# 	class base:
	# 		@operator
	# 		class inner:
	# 			pass

	# 	class subclass(base):
	# 		@operator
	# 		class inner(base.inner):
	# 			pass

	def __init__(self, function):
		self.baseClass = function

	def __get__(self, object, closure):
		if object is None:
			# The descriptor was accessed on the class object, probably so it
			# could be subclassed.  Return the baseClass-implementation.
			return self.baseClass

		return self.build(object)

	def build(self, *args, **kwd):
		return self.baseClass(*args, **kwd)


class memorized(action):
	# todo: this could be used to implement collection.propertyClass.
	# Note: the actual memorized value object is on THIS object.

	def __init__(self, name, *args, **kwd):
		action.__init__(self, *args, **kwd)
		self.name = '__' + name

	def __call__(self, object):
		try: return getattr(object, self.name)
		except AttributeError:
			o = action.__call__(self, object)
			setattr(object, self.name, o)
			return o

	class property(property):
		def __init__(self, object):
			self.memory = memorized(object.__name__, object) # auto
		def __get__(self, object, closure):
			return self.memory()


class memorizedOperator:
	'''
	@operator.memorized('interface')
	class interface(network):
		pass

		'''

	def __init__(self, name):
		self.name = name
	def __call__(self, object):
		return self.memorization(self.name, object)

	class memorization(operator):
		def __init__(self, name, object):
			operator.__init__(self, object)
			self.build = memorized(name, self.build)
	named = memorization


	@classmethod
	def auto(self, object):
		'''
		@operator.memorized.auto
		class interface(network):
			pass

			'''

		return self.memorization(object.__name__, object)


operator.memorized = memorizedOperator


class propertyAlias(property):
	# move into .arch
	def __init__(self, name):
		self.name = name
	def __get__(self, object, closure):
		return getattr(object, self.name)


class eventAccessor(scopeObject):
	# todo: collection?
	def __init__(self, object):
		self.object = object

	def __getattr__(self, name):
		try: return object.__getattribute__(self, name)
		except AttributeError:
			ev = delegate(self.object) # Note: does not re-reference this.
			setattr(self, name, ev)
			return ev

# XXX setting eventAccessor.operator works, but gettng it
# causes it to call the operator(property).__get__ method... ?!?
eventAccessorOperator = operator(memorized('eventAccessor', eventAccessor))


class structure:
	@classmethod
	def loadOverlay(self, o, m):
		if isinstance(m, dict):
			for (name, value) in m.items():
				if isinstance(value, ClassType):
					if issubclass(value, actionable.command):
						value(o) # register.
					else:
						self.loadOverlay(getattr(o, name), dictOf(value)) # build network.

				elif isinstance(value, dict):
					self.loadOverlay(getattr(o, name), value)
				else:
					setattr(o, name, value)


class network(list, connective, structure): # entwork
	# generic extensible, symbolic, multiplexed network interface

	# Serves as an object that can have mapped attributes, or
	# attributes that automatically resolve to sub-networks,
	# or functions like an event-observer dispatch model.
	# And, of course it can be subclassed.
	#
	# todo: implement extension/manifold..?
	#
	# Document Object Model Pattern:
	#   each network instance represents an element, because it
	#   has mapped attributes, and child (instance) attributes,
	#   and, it functions as an event listener.  Parent elements
	#   are represented by the 'object' compositional attribute,
	#   the root parent as 'source', and it natively allows
	#   method extension because it's a class.  When compiled
	#   with the operator property decorator, entities can
	#   easily define classes of network interfaces.
	#
	event = event

	def __init__(self, object, *args, **kwd):
		list.__init__(self, args)
		self.object = object
		self.mapping = dict(kwd)
		self.loadOverlay(self, dictOf(classOf(self))) # move into init?
		self.init()


	def init(self):
		pass


	@property
	def source(self):
		o = self
		while isinstance(o, network):
			o = o.object

		return o

	@property
	def subnetworks(self):
		r = list(self.__dict__.keys())
		r.remove('object')
		r.remove('mapping')
		return [s for s in r if not s.startswith('_')]

	def __repr__(self):
		return '<%s: %d subnetworks, %d keys, %d listeners>' % \
				(self.__class__.__name__, len(self.subnetworks),
				 len(self.mapping), len(self))

	def __getattr__(self, name):
		try: return object.__getattr__(self, name)
		except AttributeError:
			i = network(self) # Note: Instantiate basic network type, not subclass-type.
			setattr(self, name, i)
			return i

	def __getitem__(self, name):
		if isinstance(name, int):
			return list.__getitem__(self, name)

		return self.mapping[name]

	def __setitem__(self, name, value):
		self.mapping[name] = value

	def __iadd__(self, function):
		self.append(function)
		return self

	@contextmanager
	def context(self, function):
		self += function
		try: yield
		finally: self -= function

	def __call__(self, *args, **kwd):
		for i in self:
			i(self, *args, **kwd)


# Although, this will generally be customized, it should be extended
# to include structural configuration.  So... it's not actually useful.
interfaceOperator = operator(memorized('interface', network))


class actionable(network):
	# A verb (command-line) interactivity base class. 
	# todo: this "conflicts" with named core type.  So as a spatial.architecture,
	# it is still qualified different, by the problem is that the name 'action'
	# is appropriate for this kind of behavior, and it is 'able' because it is
	# basically a base class...  So it means that if you want to build with it,
	# but also use the core type, you're going to have to be aware until I can
	# come up with a more suitable name.

	'''
	extending:

	class rpcAction(actionable):
		class command(actionable.command):
			@classmethod
			def parse(self, input):
				(verb, args, kwd) = loadable(input).loading.json
				return self.implementation(verb, *args, **kwd)

	class object:
		'object().rpc("[method, [], {}]")'

		def do(self, *args, **kwd):
			pass

		@operator
		class rpc(rpcAction):
			def init(self):
				self.method(self) # Register method command.
				self.methods['do'] = self.object.do

			class method(rpcAction.command):
				name = 'method'

				class implementation(rpcAction.command.verb):
					def __init__(self, name, *args, **kwd):
						self.name = name
						self.args = args
						self.kwd = kwd

					def __call__(self, actor, *args, **kwd):
						actor.interface.methods[self.name] \
							(*self.args, **self.kwd)


		'''

	class NotFound(Exception):
		pass
	class Yield(Exception):
		pass

	def __call__(self, interface, input, *args, **kwd):
		# Currently this controls the command-implementation-and-dispatch lifecycle
		# for a particular network interface mapping.
		#
		# todo: Override network invocator to provide event-oriented dispatch
		# (that is, the event itself defines the invocation).
		#
		# To provide another input type, subclass the 'command' class on the interface,
		# and override the parse methods.

		verb = self.command.parseActionVerb(interface, input)
		return verb(self.object, *args, **kwd)


	class interpretable:
		nonalpha = '!"#$%&\'()*+,-./0123456789:;<=>?@[\\]^_`{|}~'

		@classmethod
		def parseCommandArgs(self, command):
			# command = command.lstrip()
			if command:
				if command[0] in self.nonalpha:
					return (command[0], command[1:])

				i = command.find(' ')
				if i > 0:
					return (command[:i], command[i:])

				return (command, '')

			return ('', '')


		parseCommand = parseCommandArgs


		class unparseable(Exception):
			def __init__(self, command):
				self.command = command


		@classmethod
		def parse(self, command):
			# This is kind of dumb: but, it is used to initiating command events.
			if isinstance(command, str):
				command = command.lstrip()
				if command:
					return self.parseCommand(command)

			raise self.unparseable(command)


		@classmethod
		def parseActionVerb(self, interface, command):
			# This implementation defines the input type (text line string).
			# Pre-parse the command into parts by relying on base command type.
			(verb, argstr) = self.parse(command)

			# Search for this command implementation in the network.
			try: action = interface[verb]
			except KeyError:
				raise actionable.NotFound(verb)

			# Build the interface-command-specific verb implementation.
			# The action is now an polymorphic instance of the command type.
			return action.implementation(action, command, verb, argstr)


	class command(interpretable):
		def __init__(self, interface):
			interface[self.name] = self

		def parseSubcommandArgs(self, verb, *args, **kwd):
			# Also gives an opportunity to rebind the verb instance.
			return verb.parseArgs(*args, **kwd)

		def __call__(self, verb, actor, *args, **kwd):
			verb = self.parseSubcommandArgs(actor, *args, **kwd)
			return verb(actor, *args, **kwd) # and pass *args, **kwd?


		class verb(network.event):
			def __init__(self, action, command, verb, argstr):
				self.action = action
				self.command = command
				self.verb = verb
				self.argstr = argstr

			def parseArgs(self, *args, **kwd):
				# object, subject, direct object, preposition, etc.
				self.args = self.argstr.split() # shlex.split?
				return self

			def __call__(self, *args, **kwd):
				return self.action(self, *args, **kwd)

			def __str__(self):
				return '<%s: %s>' % (self.__class__.__name__, self.verb)


		class define:
			'''
			@operator
			class interface(action):
				def init(self):
					@actionable.command.define('look', self)
					def lookCommand(actor, verb):
						pass

						'''

			def __init__(self, name, interface = None):
				self.name = name
				self.interface = interface

			def __call__(self, function):
				o = definition(self.name, function)
				i = self.interface

				if i is not None:
					return o(o)

				return o


			class definition:
				def __init__(self, name, function):
					self.name = name
					self.function = function

				def __call__(self, network):
					# Registration.
					return call(classObject(self.name, actionable.command,
											implementation = self.implementation,
											name = self.name), network)

				@property
				def implementation(self):
					from new import instancemethod as bind
					return classObject('implementation', action.command.verb,
									   __call__ = bind(self.run, self))

				def run(self, verb, actor, *args, **kwd):
					return self.function(actor, verb) # Note: rearrangement of pos args.



	# (any source) initiate location change
	#	 start virtual method for begin change
	#	 on completion, perform location transfer
	#	 fire location-transfer event
	#	 on completion of event, call virtual machine for change completion
	#		 or, resume existing task-call as completion.
	#
	# "(any source)" must wait for completion, which means it must pass
	# a callback to location-change routine, or, it must somehow state
	# conditional invariance (?) by asserting control over the operation.
	#
	# transitively atomic operation

	# def resumeVMCall(frame, result):
	# 	from stuphos.kernel import resumeTask
	# 	resumeTask(frame.task, value = result)

	# class chain(list):
	# 	def __init__(self, *functions):
	# 		list.__init__(self, functions)

	# 	def __call__(self, *args, **kwd):
	# 		e = []

	# 		for o in self:
	# 			e = o(*((e,) + arg), **kwd)

	# 		return e


	# class entity:
	# 	def locationChangeComplete(self, frame, location):
	# 		# Stylized as native.
	# 		evCall = deferred(self.events.locationChangeComplete,
	# 						  location = location)

	# 		evCall = chain(evCall, deferred(resumeVMCall, frame))


class delegate(list):
	# Broadcasting listener.
	# todo: verb reactions may be stored on the object, but they're
	# activated by another entity.
	# Note: this is not necessary because of 'network', but the 'invocation'
	# event-dispatch-handling object is useful.
	def __init__(self, object = None):
		self.object = object

	def listen(self, callback):
		self.append(callback)
		return self

	def unlisten(self, callback):
		self.remove(callback)
		return self

	def __call__(self, *args, **kwd):
		return self.invocation(self).dispatch(*args, **kwd)

	class invocation(list):
		# todo: derive from chain?
		def __init__(self, event):
			self.event = event

		def dispatch(self, *args, **kwd):
			for callback in self:
				self.append(self.dispatchOne(callback, *args, **kwd))

			return self

		def dispatchOne(self, callback, *args, **kwd):
			return callback(self, *args, **kwd)



class collection(list):
	# @operator
	# class collection(list):
	# 	pass

	# class entities(collection):
	# 	pass

	def __init__(self, object):
		self.object = object

	def mapIterator(self, function, *args, **kwd):
		for i in self:
			yield function(i, *args, **kwd) # pass self?
	each = mapIterator
	# map = listing(each)

	def find(self, function):
		for i in self:
			if function(i):
				yield i

	def __iadd__(self, objects):
		self.extend(objects)
		return self
	__add__ = __iadd__


	@classmethod
	def component(self, object):
		# Bind this class derivation to property derivation.
		# Return a property object accessing this class derivation.
		i = subClass(object.propertyClass,
					 collectionClass = object) # XXX shouldn't this be self?

		object.propertyImpl = i
		return i(object.__name__)

	class propertyClass(property):
		def __init__(self, name):
			self.name = name

		def __getitem__(self, object):
			name = '__%s' % self.name
			try: return getattr(object, name)
			except AttributeError:
				c = self.collectionClass(object)
				setattr(object, name, c)
				return c

		def __get__(self, object, closure):
			return self[object]
		def __set__(self, object, value):
			pass


def setAttribute(object, name, attrClass, *args, **kwd):
	try: return getattr(object, name)
	except AttributeError:
		i = attrClass(*args, **kwd)
		setattr(object, name, i)
		return i


class inheritable:
	# Noticeably, this doesn't do much but implement spherical(), search(), and extend shell (and namespace).
	# It shares a hierarchy like that of network.

	def spherical(self, name):
		# Relative attribute.
		try: return getattr(self, name)
		except AttributeError as e:
			if self.location is None:
				raise e

			return self.sphere.spherical(name)


	@property
	def containerization(self):
		s = self

		while s is not None:
			yield s

			try: s = s.sphere
			except AttributeError:
				break


	@property
	def outermost(self):
		s = self

		try:
			while True:
				s = s.sphere

		except AttributeError:
			pass

		return s


	def search(self, callable):
		if isinstance(callable, str):
			def callable(e, name = callable):
				return e.name == name

		for e in self.entities:
			if callable(e):
				yield e


	# todo: what about an 'inheritable' network interface -- manifold?
	# this already kind of exists because of network layering: bind a
	# new network layer, and then access its 'source' property:
	#
	# class object:
	#     'object().interface.extension["source"](...)'

	#     @operator
	#     class interface(network):
	#         @operator
	#         class extension(network):
	#             def init(self):
	#                 # In case anyone cares.. map the item to the interface.
	#                 self.object['source'] = self.source.interface
	#                 self += self.callback

	#                 # Install a new view on the original object.
	#                 self.action = self.behavior(self.source)

	#             class behavior(network):
	#             	'object().interface.extension.action("say Performing command!")'

	#             	def __call__(self, *args, **kwd):
	#             		# Bound to the original source object.
	#             		self.object.interpretCommand(*args, **kwd)

	#             def callback(self, *args, **kwd):
	#                 pass


class graph(list):
	# An interactive view -- an object API.
	# see network

	# This generalizes function call values based on its arguments,
	# which can be either dynamic (programmatic) or simplistic
	# (attribute or index resolution).

	def __init__(self, object):
		self.object = object
		self.namespace = dict()


	class resolution:
		'''
		Formulates a graph lookup by providing a description framework.
		A resolution is a context-sensitive operation defined by the descriptor.

		'''

		def __init__(self, graph):
			self.graph = graph


		class edge:
			pass
		class link(edge):
			pass

		class descriptor(link):
			'Knows how to look up a value in a graph based on another value.'

			# Different interface types:
			#   an object with member methods
			#   an object with members that point to callable objects
			#   an object with members that have members that point to methods
			#   an object with indices that are callable
			#   an object with operators
			#   ...

			def __init__(self, value):
				self.value = value

			@contextmanager
			def __call__(self, graph):
				yield self[graph]

		class string(descriptor):
			def __getitem__(self, graph):
				return graph.namespace[self.value]

		class index(descriptor):
			def __getitem__(self, graph):
				try: return graph.namespace[self.value]
				except KeyError:
					return graph[self.value]

		class sequence(descriptor):
			def __getitem__(self, graph):
				raise NotImplementedError

		def __getitem__(self, o):
			'Maps a description to a graph-resolution operation.'

			if isinstance(o, str):
				return self.string(o)
			if isinstance(o, int):
				return self.index(o)
			if isinstance(o, (list, tuple)):
				return self.sequence(o)

			return o

		class NotFound(KeyError):
			def __init__(self, graph, descriptor):
				self.graph = graph
				self.descriptor = descriptor

		def __getitem__(self, descriptor):
			'Determine graph action lookup method based on description.'

			method = self[descriptor](self.graph)
			if callable(method):
				return method

			raise self.NotFound(self.graph, descriptor)


	def resolve(self, descriptor):
		# A graph could just implement this method to do a simple action lookup.
		# Or, you could implement a more complex resolution description:
		#
		# @apply
		# class lookup(graph.resolution.descriptor):
		#     def __getitem__(self, graph):
		#         return graph(['path', 'to', 'resource', 'method'])
		#
		# i.graph(lookup())

		return self.resolution(self)[descriptor]

	def __call__(self, descriptor, *args, **kwd):
		with self.resolve(descriptor) as method:
			return method(self.object, *args, **kwd)



from . import CommandMessage, AsCommandClass
from .. import deploy # xxx dependence
deploymentApi = runtime.MUD.Deployment.API

@AsCommandClass('application/x-stuph-deployment')
class DeploymentCommandMessage(CommandMessage):
    MULTITHREAD = True

    # (Something like this)
    # Actually, while a material resource is one good way to provide the archive,
    #   it is still an in-memory construct:
    #
    #   A deployment archive requires extraction to a directory.
    #   Also allow for this in-memory buffer to come from the command payload.
    #
    class MaterialResourceArchive(runtime[deploymentApi].getArchiveBaseClass()):
        def extractTo(this):
            pass

    def Execute(this):
        runtime[deploymentApi].LiveDeploymentUrlFromConfigAndArchiveOnMultithread \
            (this.MULTITHREAD, this['X-Deployment-Config'],
             this.MaterialResourceArchive(), # todo: with payload
             this.get('X-Deployment-Update-Message', ''))




from contextlib import contextmanager
from types import GeneratorType as generator

def withAs(ctx):
	with ctx as f:
		return f

@contextmanager
def instdict(inst):
	'Create a context variable access out of the variable dictionary.'
	yield inst.__dict__

class Environment(object):
	from new import instance

	class Options:
		from pprint import pformat as pf
		pf = staticmethod(pf)

		def __repr__(self):
			return '<Environment.Options %s>' % self.pf(self.__dict__)

	def __init__(self, avatar, attr = ('env', 'e'), **options):
		self.avatar = avatar

		self.__options = options = dict(options)
		self.options = self.o = self.instance(self.Options, options)

		if type(attr) in (list, tuple, generator):
			for a in attr:
				setattr(avatar, a, self)

		elif type(attr) is str:
			setattr(avatar, attr, self)

	def __getattr__(self, name):
		return self.avatar.find(name, **self.__options)

	def __repr__(self):
		return '<Environment lookup %r>' % self.avatar.find

	def presearch(self, *args):
		'Presearch the environment.'
		return dict((k, getattr(self, k)) for k in args)

	@property
	@contextmanager
	def __context__(self):
		'''
		with mobile.e as e:
			print e.fido, e.wyrm, e.zifnab

		with instdict(mobile.e.presearch(*args)) as p:
			print p # The presearched dictionary result.

		'''
		yield self

ment = Environment



'''
level, name, position
social
subcmd
'''

class Compare:
    def __init__(self, value):
        self.value = value

    def compare(self, this):
        raise NotImplementedError

class CompareAttributes(Compare):
    def __init__(self, value, attr):
        self.value = value
        self.attr = attr

    def compare(self, this):
        return self.value.compare(getattr(this, self.attr))

class AndAll(Compare):
    def __init__(self, *these):
        self.these = these
    def compare(self, this):
        return all(t.compare(this) for t in self.these)

class Equals(Compare):
    def compare(self, this):
        return this == self.value

class IsSameObject(Compare):
    def compare(self, this):
        return this is self.value

class IsNotSameObject(Compare):
    def compare(self, this):
        return this is not self.value

class NotEqual(Compare):
    def compare(self, this):
        return this != self.value

Not = NotEqual

class GreaterThan(Compare):
    def compare(self, this):
        return this > self.value
class GreaterThanOrEqual(Compare):
    def compare(self, this):
        return this >= self.value

class LessThan(Compare):
    def compare(self, this):
        return this < self.value
class LessThanOrEqual(Compare):
    def compare(self, this):
        return this <= self.value

class InRange(Compare):
    def compare(self, this):
        return this in self.value

class Startswith(Compare):
    def compare(self, this):
        return this.startswith(self.value)
class Endswith(Compare):
    def compare(self, this):
        return this.endswith(self.value)

def selectEntities(entities, **kwd):
    def getComparison(arg):
        if isinstance(arg, Compare):
            return arg

        if isinstance(arg, (int, str, float)):
            return Equals(arg)
        if isinstance(arg, (list, tuple)):
            return InRange(arg)

        return IsSameObject(arg)

    criteria = [CompareAttributes(getComparison(v), n) \
                for (n, v) in kwd.items()]

    criteria = AndAll(*criteria)
    criteria = criteria.compare

    for e in entities:
        if criteria(e):
            yield e

class Criteria(dict):
    def query(self, entities):
        return selectEntities(entities, **self)

class CommandQuery(Criteria):
    def query(self):
        import world
        return Criteria.query(self, world.iterent(world.command))
    __iter__ = query

    def show(self, results):
        yield 'Name            Level       Position        Subcmd   Social'
        yield '==========================================================='

        results = list(results)
        for c in results:
            yield '%-15s %-11s %-15s %-8s %-6s' % \
                  (c.name, c.level, c.position, c.subcmd, c.social)

        yield ''
        yield 'Results: %d' % len(results)
        yield ''

    def page(self, peer):
        peer.page('\n'.join(self.show(self.query())))


import re
COMPARISON_PATTERN = re.compile(r'^(name|level|position|subcmd|social)\s*' + \
                                r'(\<|\>|\<\=|\>\=|\=|\!\=|\:in\:|\:startswith\:|\:endswith\:|\:is\:|\:isnot\:)\s*' + \
                                r'(.*)$')

COMPARISON_MAP = {'<'            :LessThan,
                  '>'            :GreaterThan,
                  '<='           :LessThanOrEqual,
                  '>='           :GreaterThanOrEqual,
                  '='            :Equals,
                  '!='           :NotEqual,
                  ':in:'         :InRange,
                  ':startswith:' :Startswith,
                  ':endswith:'   :Endswith,
                  ':is:'         :IsSameObject,
                  ':isnot:'      :IsNotSameObject}

from stuphmud.server.player import ACMDLN, Option
@ACMDLN('commands-', Option('--format'),
                     Option('--count', action = 'store_true'),
                     Option('--count-only', action = 'store_true'),
                     shlex = True)
def doCommandsEx(peer, command):
    # Ignores RESERVED
    criteria = dict()
    for arg in command.args:
        m = COMPARISON_PATTERN.match(arg)
        if m is not None:
            (n, o, a) = m.groups()

            la = a.lower()
            if la == 'true':
                a = True
            elif la == 'false':
                a = False
            elif a.isdigit():
                a = int(a)
            elif a[0] in '[({':
                from json import loads
                a = loads(a)

            criteria[n] = COMPARISON_MAP[o](a)

    q = CommandQuery(**criteria)
    if command.options.format or command.options.count or command.options.count_only:
        def _(r):
            if not command.options.count_only:
                f = command.options.format.format
                for c in r:
                    yield f(command = c)

            if command.options.count or command.options.count_only:
                yield 'Results: %d' % len(r)

        peer.page('\n'.join(_(list(q))))
    else:
        q.page(peer)

    return True



#!/usr/bin/python
from .scan import *
from .show import *

from struct import pack

try:
	from md5 import new as MD5
	def checksum(x, MD5=__import__('md5').new):
		return MD5(str(x)).hexdigest()

except ImportError:
	checksum=lambda x:id(x)

def repack_item(item):
	#print `structfmt_obj_file_elem`, len(item), [type(v) for v in item]
	return pack(structfmt_obj_file_elem, *item)

# An EQ-Set is a (non-sorted) collection of items in their wear-positions
# as expressed in the final obj_file_elem parts of a rent-file.
#
# (A Rent-ID is sorted, specifically, and computed with a more appropriate
#  algorithm; see below.)
def repack_eqset(eqset):
	"Repack the EQ-Set into a buffer."
	return ''.join([repack_item(item) for item in eqset])

# This is mainly useful for re-writing the rentfile
def repack_rent(rent, result=[]):
	"""
	Reconstruct rent-file given the rent-tuple format:
		[(rent_info), (obj_file_elem) x N, ...]
	"""
	result  = [pack(structfmt_rent_info, *rent[0])]
	result += repack_eqset(rent[1:])

	return ''.join(result)

# A Rent-ID is meant to uniquely identify set of equipment,
# and is generally computed given 
#   - Sorted EQ-Set
#   - MD5 checksum the value of this written buffer
#   - Integrated secure (cryptographic) key
def compute_rentid(eqset):
	# Generate packed formats:
	eqset=list(map(repack_item, eqset))

	# Generate map for sorting, discarding old structures.  Checksums
	# are calculated for each blob used for sorting comparisons:
	eqset=list(zip(eqset, list(map(checksum, eqset))))

	# Initial EQ-Set, and sort map is abandoned for a sorted set of
	# item buffers:
	eqset.sort(lambda x, y:cmp(x[1], y[1]))
	eqset=[item[0] for item in eqset]

	# Obtain packed buffer format (skipping repack_eqset).
	# Note that this is a sorted representation, and that's
	# where we get our identification properties from::
	eqset=''.join(eqset)

	# Calculate (checksum) Rent-ID from eqbuf:
	return checksum(eqset)

def apply_item_transform(items, proc, *args, **kwd):
	return [proc(i, *args, **kwd) for i in items]

def apply_rent_script(rent, info_proc=None, item_proc=None):
	items=rent[1:]
	rent=rent[0]

	if info_proc:
		rent=info_proc(rent)

	if item_proc:
		items=[list(i) for i in items]
		items=apply_item_transform(items, item_proc)
		items=[_f for _f in items if _f]
		items=tuple([tuple(i) for i in items])

	return repack_rent((rent,) + items)

def NewDerivedFile(f):
	return f + '.out'

def path_applyscript(path, info_proc=None, item_proc=None, **kwd):
	# Keyword args: info_proc, item_proc, recursive_policy, NewDerivedFile
	locals().update(kwd)

	assert info_proc or item_proc, Error('No transformation..?')

	# XXX Test path for directory:

	outf=open(NewDerivedFile(path), 'w')
	outf.write(apply_rent_script(get_rent(path), info_proc, item_proc))
	outf.flush()
	outf.close()

# Classes of transformation applied by 
class Apply:
	def FileScript(path):
		code=open(path).read()
		code=compile(code, path, 'exec')

		def ProcessItemWithScript(item, code=code):
			ns={'item':item}
			exec(code, ns)
			return ns.get('item', item)

		return ProcessItemWithScript
	FileScript=staticmethod(FileScript)

	def ForAll(item):
		return True
	ForAll=staticmethod(ForAll)

	def ForNone(item):
		return False
	ForNone=staticmethod(ForNone)

	class IndexCheck:
		def __init__(self, index, acptbl=None):
			self.index=index
			self.acceptable=acptbl

		def IsAcceptable(self, item):
			val=item[self.index]
			actbl=self.actbl

			if type(actbl) in (list, tuple):
				return val in actbl

			elif callable(actbl):
				return actbl(val)

			return val == actbl

		__call__=IsAcceptable

	class Condition:
		def __init__(self, condition, procedure):
			self.cond=condition
			self.proc=procedure

		def TransformUnderCondition(self, item):
			if self.cond(item):
				item=self.proc(item)

			return item

		__call__=TransformUnderCondition

	# class CookedConditions(Condition):
	#	def __init__(*args):
	#		Condition.__init__(*args)
	#	def CheckUnderCookedConditions(self, item):
	#		...
	#		return Condition.TransformUnderCondition(self, item)
	#	__call__=CheckUnderCookedConditions

	class Aggregate:
		"""
		VR_NDX=0 # Virtual Number Index
		WP_NDX=4 # Wear-positional index

		vnums=range(*input(' ZBottom, ZTop? '))

		a_aggr=Apply.Aggregate()
		a_acpt=Apply.Condition(IndexCheck(VR_NDX, vnums), a_aggr)

		def PutIntoInventory(item):
			item=list(item)
			item[WP_NDX]=0

			return tuple(item)

		a_aggr.add(PutIntoInventory)
		# a_aggr.add()

		apply_objfile_script('ada.objs', item_proc=a_acpt)
		"""
		def __init__(self, sequence=[], *args, **kwd):
			self.sequence=sequence or []
			self.args=args
			self.kwd=kwd

		def ExecuteSequence(self, item):
			for s in self.sequence:
				if not item:
					break

				item=s(item, *self.args, **self.kwd)

			return item
		__call__=ExecuteSequence

		def addToSequence(self, proc):
			if type(proc) in (list, tuple):
				for p in proc:
					self.addToSequence(p)

			elif callable(proc):
				self.sequence.append(proc)

			# Method-Chaining:
			# return self
		add=__iadd__=addToSequence

def path_applymodule(path, module, *args, **kwd):
	"""
	Load module as: module, path-name, file, single-function
	Load rent file from: path

	Build item transformation application with: module
	Call path_applyscript with: path, transformation application
	"""
	locals().update(kwd)

	# XXX Move this module building into an Application routine!!

	tp=type(module)
	if tp is __import__('types').ModuleType:
		module=list(module.__dict__.values())

	elif tp is str:
		modpath=module
		module=open(module)

	elif tp is file:
		modpath=module.name

	elif callable(module):
		module=[module]

	if type(module) is file:
		ns={'__file__':modpath}
		exec(compile(module.read(), modpath, 'exec'), ns)

		# XXX pname ..??
		pname='process_item'
		isProcessItem=lambda f, pn=pname:f.__code__.co_name == pn
		module=[v for v in list(ns.values()) if isProcessItem(v)]

	assert type(module) is list

	from types import FunctionType as function
	module=[f for f in module if type(f) is function]

	return path_applyscript(path, item_proc=Apply.Aggregate(module), **kwd)

def transform_interface(path):
	args=[]
	kwd={}

	# XXX Use a menu to inquire: module path-name/object, args/kwd environment
	module=eval(input(' Module? '))

	return path_applymodule(path, module, *args, **kwd)

def edit_transformation(path, names=('transform', 'transform_item', 'process', 'process_item')):
	from nano import edit as nano

	ns={}
	code=nano()
	code=compile(code, 'fast', 'exec')
	exec(code, ns)

	from types import FunctionType as function
	module=[]
	kwd={}

	for k, v in ns.items():
		if type(v) is function and v.__code__.co_name in names:
			module.append(v)

		else:
			kwd[k]=v

	return path_applymodule(path, module, **kwd)
edittrans=edit_transformation

def fast_transformation(path, script=None):
	assert script
	return path_applymodule(path, Apply.FileScript(script))
fasttrans=fast_transformation

def show_rentid(path):
	"Compute Rent-ID given rent-file path name."
	rent=get_rent(path)
	eqset=rent[1:]

	rentid=compute_rentid(eqset)

	print(path, ':', rentid)
	for line in list_rent_info(rent[0]):
		print('    ', line)

if __name__=='__main__':
	config={}
	proc=[show_rentid]

	for a in __import__('sys').argv[1:]:
		op=a[0]
		if op in '-=+':
			parts=a[1:].split(':')
			name=parts[0]
			parts=parts[1:]

			if name in ('t', 'proc', 'procedure'):
				proc=[eval(p) for p in parts]

			elif op == '=':
				config[name]=[eval(p) for p in parts]

			else:
				config[name]=':'.join(parts)

		else:
			for p in proc:
				p(a, **config)
